/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   This file contains an implementation of an address book.

*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#include "address.h"
#include "raddress.h"
#include "error.h"
#include "file.h"
#include "xmalloc.h"
#include "ask.h"
#include "cmd.h"
#include "mail.h"
#include "select.h"
#include "folder.h"
#include "line.h"
#include "read.h"
#include "gettext.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define ABOOK_WIDTH 80

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static char       *afile = NULL;
static raddress_t *abook = NULL;

static select_t   *add_select   = NULL;
static raddress_t *add_raddress = NULL;

static select_t   *abook_select = NULL;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static void
process_line (char *line)
{
  static int   state = 0;
  static char *name  = NULL;
  static int   flags = 0;

  char      *email;
  address_t *address;

  while (*line && !isalnum (*line))
    line++;

  if (*line == '\0')
    return;
  
  switch (state){

    case 0:
      flags = strtol (line, &line, 16);
      line++;
      name  = xstrdup (line);
      break;

    case 1:
      email   = xstrdup (line);
      address = address_construct (email, name, flags);
      raddress_add (abook, address);
      xfree (email);
      xfree (name);
      email = name = NULL;
      break;

    default:
      error_ (0, 0, _("congratulations! you have just found a bug"));
      return;
  }
  state = 1 - state;
}




static void
write_address (FILE *fp, address_t *addr)
{
  fprintf (fp, "0x%X %s\n%s\n", addr->flags, addr->name, addr->email);
}



static int
addr_cmp (const void *a, const void *b)
{
  const address_t **aa = (const address_t **) a;
  const address_t **bb = (const address_t **) b;
  char             *aname;
  char             *bname;
  int               ret;
  
  aname = ((*aa)->last) ? (*aa)->last : (*aa)->name;
  bname = ((*bb)->last) ? (*bb)->last : (*bb)->name;

  ret = strcoll (aname, bname);
  return ret;
}




static char *
make_item (line_t *line, address_t *addr)
{
  line_clear (line);
  line_add_char (line, addr->flags & 0xff);
  line_add_char (line, ' ');
  line_add_str_pad (line, addr->name, 30);
  line_add_char (line, ' ');
  line_add_str (line, addr->email);

  return xstrdup (line->str);
}



static rstring_t *
make_items (void)
{
  address_t **addr;
  rstring_t  *result = rstring_create_size (abook->count + 1);
  line_t     *line   = line_create (ABOOK_WIDTH - 2 + 1);

  for (addr = abook->array; *addr; addr++){
    rstring_add (result, make_item (line, *addr));
  }
  line_destroy (line);
  return result;
}



/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

void
abook_start (void)
{
  char *start;
  char *end;
  char *buffer;
  int   size;
  FILE *fp;

  afile = ask_noninteractive ("addressbook");
  if (afile == NULL)
    afile = file_with_dir (getenv ("HOME"), ".addressbook");
  else
    afile = xstrdup (afile);
  
  fp = fopen (afile, "r");
  if (!fp){
    error_ (0, errno, _("opening addressbook: %s"), afile);
    return;
  }
  if (file_whole (fp, &buffer, &size)){
    error_ (0, errno, _("opening addressbook: %s"), afile);
    return;
  }
  fclose (fp);

  abook = raddress_create ();
  
  end = buffer;

  if (size!=0) {
    while (1) {
      start = end;
      end   = strchr (start, '\n');
      if (end == NULL)
        break;
      *end  = '\0';
      process_line (start);
      *end  = '\n';
      end++;
    }

    process_line (start);
  }

  file_free_buffer (buffer, size);
}



void
abook_sort (void)
{
  if (abook == NULL)
    return;
  qsort (abook->array, abook->count, sizeof (address_t *), addr_cmp);
}



void
abook_save (void)
{
  FILE       *fp;
  address_t **addr;
  
  abook_sort ();

  fp = fopen (afile, "w");
  for (addr = abook->array; *addr; addr++){
    write_address (fp, *addr);
  }
  fclose (fp);
}



void
abook_end (void)
{
  if (afile == NULL)
    return;

  if (abook == NULL){
    xfree (afile);
    afile = NULL;
    return;
  }
  
  abook_save ();

  if (abook){
    raddress_destroy (abook);
    abook = NULL;
  }
  
  if (afile){
    xfree (afile);
    afile = NULL;
  }
}



void
abook_open (void)
{
  rstring_t *items;

  items = make_items ();

  abook_select = select_open (items, ABOOK_WIDTH, 20, 1, 0, _("addressbook"));
  if (abook_select == NULL)
    return;

  select_show (abook_select);
  cmd_state_push (CMD_ABOOK);
}



void
abook_close (void)
{
  rstring_free_strings (abook_select->items);
  rstring_delete (abook_select->items);
  select_close (abook_select);
  cmd_state_pop ();
}



void
abook_next (void)
{
  select_next (abook_select);
}



void
abook_prev (void)
{
  select_prev (abook_select);
}



void
abook_next_page (void)
{
  select_next_page (abook_select);
}



void
abook_prev_page (void)
{
  select_prev_page (abook_select);
}



void
abook_first (void)
{
  select_first (abook_select);
}



void
abook_last (void)
{
  select_last (abook_select);
}



void
abook_remove (void)
{
  raddress_remove (abook, abook_select->bar_pos);
  select_remove_selected (abook_select);
}




void
abook_insert (void)
{
  char      *name;
  char      *email;
  char      *flags;
  address_t *addr;
  char      *item;
  line_t    *line = line_create (ABOOK_WIDTH + 2 - 1);

  name  = xstrdup (read_argument ("name: ", NULL));
  if (name == NULL || *name == '\0'){
    if (name)
      xfree (name);
    line_destroy (line);
    return;
  }
  email = xstrdup (read_argument ("email: ", NULL));
  if (email == NULL || *email == '\0'){
    xfree (name);
    if (email)
      xfree (email);
    line_destroy (line);
    return;
  }
  flags = xstrdup (read_argument ("flag: ", NULL));
  if (flags == NULL || flags == '\0'){
    xfree (name);
    xfree (email);
    if (flags)
      xfree (flags);
    line_destroy (line);
    return;
  }

  addr = address_construct (email, name, *flags);
  item = make_item (line, addr);
  raddress_add (abook, addr);
  select_add_item (abook_select, item);

  line_destroy (line);
  xfree (name);
  xfree (email);
  xfree (flags);
}



void
abook_change_name (void)
{
  address_t  *addr = abook->array[abook_select->bar_pos];
  char       *name = read_argument ("name: ", addr->name);
  char      **item = abook_select->items->array + abook_select->bar_pos;
  line_t     *line = line_create (ABOOK_WIDTH + 2 - 1);

  if (name == NULL || *name == '\0'){
    line_destroy (line);
    return;
  }
  
  addr  = address_construct (addr->email, name, addr->flags);
  xfree (*item);
  *item = make_item (line, addr);
  abook->array[abook_select->bar_pos] = addr;
  select_redraw (abook_select);

  line_destroy (line);
}



void
abook_change_email (void)
{
  address_t  *addr  = abook->array[abook_select->bar_pos];
  char       *email = read_argument ("email: ", addr->email);
  char      **item  = abook_select->items->array + abook_select->bar_pos;
  line_t     *line  = line_create (ABOOK_WIDTH + 2 - 1);

  if (email == NULL || *email == '\0'){
    line_destroy (line);
    return;
  }
  
  addr  = address_construct (email, addr->name, addr->flags);
  xfree (*item);
  *item = make_item (line, addr);
  abook->array[abook_select->bar_pos] = addr;
  select_redraw (abook_select);

  line_destroy (line);
}




void
abook_change_flag (void)
{
  address_t  *addr  = abook->array[abook_select->bar_pos];
  char       *flags = read_argument ("flag: ", NULL);
  char      **item  = abook_select->items->array + abook_select->bar_pos;
  line_t     *line  = line_create (ABOOK_WIDTH + 2 - 1);

  if (flags == NULL || *flags == '\0'){
    line_destroy (line);
    return;
  }
  
  addr->flags = *flags;
  xfree (*item);
  *item       = make_item (line, addr);

  select_redraw (abook_select);
  line_destroy (line);
}
  



void
abook_add_open (void)
{
  rstring_t *items;
  mail_t    *mail;
  
  mail = mail_array + folder_bar_index ();
  
  add_raddress = raddress_create ();

  if (mail->from)
    raddress_add (add_raddress, mail->from);
  if (mail->reply_to && mail->reply_to != mail->from)
    raddress_add (add_raddress, mail->reply_to);
  if (mail->to)
    raddress_add_array (add_raddress, mail->to);

  items = raddress_to_strings (add_raddress);

  add_select = select_open (items, COLS - 20, 20, 1, 0, _("select address"));

  if (add_select == NULL)
    return;
  
  select_show (add_select);

  cmd_state_push (CMD_ABOOK_ADD);
}



void
abook_add_close (void)
{
  if (add_raddress){
    raddress_destroy (add_raddress);
    add_raddress = NULL;
  }

  rstring_delete (add_select->items);
  select_close (add_select);

  cmd_state_pop ();
}



void
abook_add_next (void)
{
  select_next (add_select);
}



void
abook_add_prev (void)
{
  select_prev (add_select);
}



void
abook_add_next_page (void)
{
  select_next_page (add_select);
}



void
abook_add_prev_page (void)
{
  select_prev_page (add_select);
}



void
abook_add_first (void)
{
  select_first (add_select);
}



void
abook_add_last (void)
{
  select_last (add_select);
}



void
abook_add_hit (void)
{
  address_t *addr = add_raddress->array[add_select->bar_pos];
  
  raddress_add (abook, addr);
}



/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE abook.c
 *
 ****************************************************************************/
