/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   Different sring manipulation routines.
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <stdarg.h>
#include <string.h>

#include "mail.h"
#include "xmalloc.h"
#include "wrapbox.h"
#include "misc.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define FIRST_ALLOC 20

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/

typedef struct {
  char *str;
  char *end;
  int   size;
} str_t;

/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/
/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static void
initialize_str (str_t *s)
{
  s->size = FIRST_ALLOC;
  s->str  = xmalloc (s->size);
  s->end  = s->str;
  *s->end = '\0';
}


static void
resize_str (str_t *s)
{
  int len = s->end - s->str;
  
  if (len == s->size - 1){
    s->size = (s->size + 1) * 2;
    s->str  = xrealloc (s->str, s->size);
    s->end  = s->str + len;
  }
}


static void
put_char (str_t *s, char c)
{
  resize_str (s);
  *s->end = c;
  s->end++;
  *s->end = '\0';
}



static void
put_string (str_t *s, const char *t)
{
  int len;
  int oldlen = s->end - s->str;
  
  if (!t)
    return;

  len = strlen (t);
  if (len + oldlen >= s->size - 1){
    s->size = (s->size + len) * 2;
    s->str  = xrealloc (s->str, s->size);
  }

  memcpy (s->str + oldlen, t, len);
  s->str[oldlen + len] = '\0';
  s->end = s->str + oldlen + len;
}



static void
put_string_to_len (str_t *s, const char *t, int len)
{
  int oldlen = s->end - s->str;
  int reallen;

  if (!t)
    return;

  reallen = strlen (t);

  if (len + oldlen >= s->size - 1){
    s->size = (s->size + len) * 2;
    s->str  = xrealloc (s->str, s->size);
  }

  if (reallen < len){
    memcpy (s->str + oldlen, t, reallen);
    memset (s->str + oldlen + reallen, ' ', len - reallen);
  }
  else {
    memcpy (s->str + oldlen, t, len);
  }

  s->str[oldlen + len] = '\0';
  s->end = s->str + oldlen + len;
}



static void
put_int_fixed (str_t *s, int d)
{
  static char buf[15];

  sprintf (buf, "%4d", d);
  put_string (s, buf);
}


static void
put_int (str_t *s, int d)
{
  static char buf[15];

  sprintf (buf, "%d", d);
  put_string (s, buf);
}


/**
 * Suppose a mail:
 *  From: Johnny Masta Killa <johnny@masta.com>
 *  Subject: Bow before me!
 *  Date: Wed, 7 May 2003 15:36:46 +0200
 *
 */
static void
put_mail_special (str_t *s, char c, mail_t *mail)
{
  int   size;
  char *date;

  resize_str (s);
  switch (c){

    case '%':
      put_char (s, '%');
      break;

      /* Johnny Masta Killa */
    case 'f':
      if (mail->from->name)
        put_string (s, mail->from->name);
      else
        put_string (s, mail->from->full);
      break;

      /* Johnny Masta Killa <johnny@masta.com> */
    case 'F':
      put_string (s, mail->from->full);
      break;

      /**
       * Johnny Masta Kil
       * this one is filled with spaces to ADDR_NAME_LEN if needed
       */
    case 'x':
      put_string_to_len (s, mail->from->name_short, ADDR_NAME_LEN);
      break;

      /* Wed, 7 May 2003 15:36:46 +0200 */
    case 'd':
      put_string (s, mail->date_str);
      break;

      /* 7 May */
    case 'D':
      date = date_string ("%b %d", mail->date);
      put_string (s, date);
      xfree (date);
      break;

      /* Bow before me! */
    case 's':
      put_string (s, mail->subject);
      break;

      /* JM */
    case 'i':
      if (mail->from->name)
        put_char (s, *mail->from->name);
      else
        put_char (s, *mail->from->full);
      if (mail->from->last)
        put_char (s, *mail->from->last);
      break;

      /* it's just an index (useful for debugging) */
    case 'I':
      put_int_fixed (s, mail->index);
      break;

      /* (3.5K) */
    case 'S':
      size = wrapbox_mail_size (mail);
      put_string (s, human_size_paren (size));
      break;

      /* these are flags */
    case '$':
      if (mail->flags & FLAG_FLAGGED)
        put_char (s, 'F');
      else if (mail->flags & FLAG_READ)
        put_char (s, ' ');
      else if (mail->flags & FLAG_OLD)
        put_char (s, 'O');
      else
        put_char (s, 'N');

      if (mail->flags & FLAG_TRASHED)
        put_char (s, 'T');
      else if (mail->flags & FLAG_ANSWERED)
        put_char (s, 'r');
      else if (mail->flags & FLAG_PASSED)
        put_char (s, 'f');
      else
        put_char (s, ' ');
      break;

      /* attachments indicator */
    case '?':
      if (mail->rmime->count > 1)
        put_char (s, '+');
      else
        put_char (s, ' ');
      break;

      /* spam indicator */
    case '#':
      if (mail->flags & FLAG_SPAM)
        put_char (s, '#');
      else if (mail->flags & FLAG_LEGITIMATE
               || mail->flags & FLAG_NOSPAM)
        put_char (s, ' ');
      else
        put_char (s, '?');
      break;
  }
}


static void
put_editor_special (str_t *s, char c, const char *fname, int lineno)
{
  resize_str (s);
  switch (c){

    case '%':
      put_char (s, '%');
      break;

    case 'f':
      put_string (s, fname);
      break;

    case 'd':
      put_int (s, lineno);
      break;
  }
}


static char *
substitute_mail (const char *fmt, mail_t *mail, char *buf)
{
  str_t       result;
  const char *seek;

  if (buf){
    result.str  = buf;
    result.end  = buf;
    result.size = 1000000000;
    *result.end = '\0';
  }
  else {
    initialize_str (&result);
  }

  for (seek = fmt; *seek; seek++){
    if (*seek == '%'){
      put_mail_special (&result, seek[1], mail);
      seek++;
    }
    else
      put_char (&result, *seek);
  }

  return result.str;
}



static char *
substitute_editor (const char *fmt, const char *fname, int lineno)
{
  str_t       result;
  const char *seek;

  initialize_str (&result);
  for (seek = fmt; *seek; seek++){
    if (*seek == '%'){
      put_editor_special (&result, seek[1], fname, lineno);
      seek++;
    }
    else
      put_char (&result, *seek);
  }

  return result.str;
}

/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

char *
eprintf_mail (const char *fmt, mail_t *mail, char *buf)
{
  char    *result;

  result = substitute_mail (fmt, mail, buf);
  return result;
}



char *
eprintf_editor (const char *fmt, const char *fname, int lineno)
{
  char *result;

  result = substitute_editor (fmt, fname, lineno);
  return result;
}

/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE eprintf.c
 *
 ****************************************************************************/
