/* 
   elmo - ELectronic Mail Operator 

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   error handling
   
*/

#define _GNU_SOURCE 1

/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <pthread.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>

#include "ecurses.h"
#include "xmalloc.h"
#include "line.h"
#include "screen.h"
#include "elmo.h"
#include "cmd.h"
#include "error.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static int is_cleared = 0;

static time_t last_report_time = 0;

static pthread_mutex_t strerror_mutex;

static WINDOW          *error_window;
static pthread_mutex_t  window_mutex;
static line_t          *error_line   = NULL;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static char *
prepare_error_msg (int errnum, const char *fmt, va_list ap)
{
  char *result;
  char *user_msg;
  int   user_len;
  char *system_msg;
  int   system_len;
  
  user_len = vasprintf (&user_msg, fmt, ap);

  if (errnum){
    pthread_mutex_lock (&strerror_mutex);
    system_msg = strerror (errnum);
    system_len = strlen (system_msg);
    result     = xmalloc (user_len + 2 + system_len + 1);
    memcpy (result, user_msg, user_len);
    memcpy (result + user_len + 2, system_msg, system_len + 1);
    pthread_mutex_unlock (&strerror_mutex);
    result[user_len]     = ':';
    result[user_len + 1] = ' ';
  }
  else {
    result = xstrdup (user_msg);
  }

  free (user_msg);
  
  return result;
}



static int
can_display (void)
{
  int result;
  
  pthread_mutex_lock (&window_mutex);
  result = time (NULL) - last_report_time > 1;
  pthread_mutex_unlock (&window_mutex);

  return result;
}



static void
display_msg (char *msg, int color, int permanent)
{
  line_put_str (error_line, msg);
  pthread_mutex_lock (&window_mutex);
  wbkgd (error_window, color);
  mvwaddstr (error_window, 0, 0, error_line->str);
  screen_touchwin (error_window);
  screen_wnoutrefresh (error_window);

  last_report_time = time (NULL);
  is_cleared       = permanent;
  pthread_mutex_unlock (&window_mutex);
}


static void
display_error_msg (char *msg)
{
  
  while (!can_display ())
    usleep (100);
  
  display_msg (msg, COLOR_PAIR (4), 0);
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

void
error_init (void)
{
  error_line   = line_create (COLS - 13);
  error_window = cmd_echo_area;
  wbkgd (error_window, COLOR_PAIR (4));
}



void
error_ (int status, int errnum, const char *fmt, ...)
{
  va_list  ap;
  char    *msg;

  va_start (ap, fmt);
  msg = prepare_error_msg (errnum, fmt, ap);
  va_end (ap);

  display_error_msg (msg);

  if (status){
    usleep (800000);
    elmo_finish (0);
    fprintf (stderr, "%s\n", msg);
    xfree (msg);
    exit (status);
  }
  xfree (msg);
}



void
error_regex (int errcode, regex_t *compiled)
{
  int   length = regerror (errcode, compiled, NULL, 0);
  char *buffer = xmalloc (length + 1);

  regerror (errcode, compiled, buffer, length);

  display_error_msg (buffer);
  xfree (buffer);
}



void
error_msg (const char *fmt, ...)
{
  va_list  ap;
  char    *msg;

  va_start (ap, fmt);
  msg = prepare_error_msg (0, fmt, ap);
  va_end (ap);

  display_msg (msg, COLOR_PAIR (7), 1);
  xfree (msg);
}



void
error_vmsg (const char *fmt, va_list ap)
{
  char *msg;

  msg = prepare_error_msg (0, fmt, ap);
  display_msg (msg, COLOR_PAIR (7), 1);
  xfree (msg);
}



void
error_free_resources (void)
{
  delwin (error_window);
  line_destroy (error_line);
}



void
error_clear_force (void)
{
  pthread_mutex_lock (&window_mutex);
  line_clear (error_line);
  mvwaddstr (error_window, 0, 0, error_line->str);
  screen_wnoutrefresh (error_window);
  is_cleared = 1;
  pthread_mutex_unlock (&window_mutex);
}



void
error_clear_old (void)
{
  int should;
  int diff;
  
  pthread_mutex_lock (&window_mutex);
  diff   = time (NULL) - last_report_time;
  should = diff > 4;
  pthread_mutex_unlock (&window_mutex);

  if (should){
    error_clear_force ();
  }
}



void
error_clear (void)
{
  if (is_cleared)
    return;
  error_clear_old ();
}

/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE error.c
 *
 ****************************************************************************/
