/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2002, 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   operations on mail array
   
*/

#define _GNU_SOURCE 1

/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <string.h>

#include "xmalloc.h"
#include "mail.h"
#include "mlex.h"
#include "debug.h"
#include "rstring.h"
#include "hash.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

/**
 * Seems to be consistent.  I expect box to have about 100 messages and
 * the hash table will have 2^7 elements which is more than 100.
 */
#define INITIAL_HASH_EXPONENT   7
#define INITIAL_MAIL_COUNT    100

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

/**
 * This table holds Message-IDs of all the message in the box.
 */
static htable_t *msgid_table = NULL;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/

mail_t   *mail_array             = NULL;
unsigned  mail_count             = 0;
unsigned  mail_allocated         = INITIAL_MAIL_COUNT;
unsigned  mail_not_replies_count = 0;
unsigned  mail_old_count         = 0;
unsigned  mail_read_count        = 0;

char *mail_mua_names[] = {
  "unknown",
  "elmo",
  "Mutt",
  "Gnus",
  "Pine",
  "ELM",
  "KMail",
  "Evolution",
  "Sylpheed",
  "Mozilla",
  "The Bat!",
  "Eudora",
  "Microsoft Outlook Express",
  "Micorsoft Outlook",
};

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/

static void allocate_resources (void);

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static void
allocate_resources (void)
{
  mail_array             = xmalloc (mail_allocated * sizeof (mail_t));
  mail_count             = 0;
  mail_not_replies_count = 0;
}



static int
find_parent_in_hash (mail_t *mail)
{
  int      i;
  entry_t *entry = NULL;

  if (mail->in_reply_to){
    for (i = mail->in_reply_to->count; i; i--){
      entry = htable_lookup (msgid_table, mail->in_reply_to->array[i - 1]);
      if (entry)
        break;
    }
  }

  if (entry){
    mail_array[(int) entry->content].child_count++;
    mail_not_replies_count--;
    return (int) entry->content;
  }
  else {
    return -1;
  }
}



static void
put_msgid_into_hash (int index)
{
  mail_t *mail = mail_array + index;
  
  if (msgid_table == NULL){
    msgid_table = htable_create (INITIAL_HASH_EXPONENT);
  }

  if (!mail->msg_id)
    return;

  /**
   * This is where we can eliminate duplicates...
   */
  htable_insert (msgid_table, mail->msg_id, (void *) index);
}



static void
destroy_int (void *a)
{}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/


void
mail_destroy (mail_t *mail)
{
  if (mail->subject)
    xfree (mail->subject);
  if (mail->msg_id)
    xfree (mail->msg_id);
  if (mail->in_reply_to){
    xfree (mail->in_reply_to->array[0]);
    rstring_delete (mail->in_reply_to);
  }
  if (mail->date_str)
    xfree (mail->date_str);
  if (mail->tree)
    xfree (mail->tree);
  if (mail->flags & FLAG_MAILDIR)
    xfree (mail->place.file_name);
  if (mail->rmime)
    rmime_destroy (mail->rmime);
  if (mail->to)
    raddress_destroy (mail->to);
  if (mail->cc)
    raddress_destroy (mail->cc);
  if (mail->bcc)
    raddress_destroy (mail->bcc);
}



void
mail_free_resources (void)
{
  mail_reset_array ();
  
  if (mail_array){
    xfree (mail_array);
    mail_array = NULL;
  }
}



void
mail_reset_array (void)
{
  int i;

  for (i = 0; i < mail_count; i++)
    mail_destroy (mail_array + i);
  if (mail_array)
    xfree (mail_array);
  mail_array             = NULL;
  mail_count             = 0;
  mail_not_replies_count = 0;
  mail_read_count        = 0;
  mail_old_count         = 0;

  if (msgid_table){
    htable_destroy (msgid_table, destroy_int);
    msgid_table = NULL;
  }
}




void
mail_set_array_size (size_t size)
{
  mail_free_resources ();
  mail_allocated = size;
  allocate_resources ();
}




mail_t *
mail_insert (mail_t *mail)
{
  if (!mail_array)
    allocate_resources ();

  if (mail_allocated == mail_count){
    mail_allocated = (mail_allocated + 1) * 2;
    mail_array     = xrealloc (mail_array, mail_allocated * sizeof (mail_t));
  }

  if (mail->flags & FLAG_READ)
    mail_read_count++;
  if (mail->flags & FLAG_OLD)
    mail_old_count++;

  memcpy (mail_array + mail_count, mail, sizeof (mail_t));
  mail_count++;
  mail_not_replies_count++;
  return mail_array + (mail_count - 1);
}



void
mail_find_parents (void)
{
  int i;

  for (i = 0; i < mail_count; i++){
    put_msgid_into_hash (i);
  }    
  
  for (i = 0; i < mail_count; i++){
    mail_array[i].parent = find_parent_in_hash (&mail_array[i]);
  }

  if (msgid_table){
    htable_destroy (msgid_table, destroy_int);
    msgid_table = NULL;
  }
}



mail_t *
mail_new_array (void)
{
  mail_t *result = xmalloc (mail_count * sizeof (mail_t));

  return result;
}



void
mail_switch_to_array (mail_t *array)
{
  xfree (mail_array);

  mail_array = array;
}



void
mail_copy_to_array (mail_t *array, unsigned oldindex, unsigned newindex)
{
  memcpy (array + newindex, mail_array + oldindex, sizeof (mail_t));
}


/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE mail.c
 *
 ****************************************************************************/
