#ifndef __MAIL_H__
#define __MAIL_H__ 1
/* 
   elmo - Mail User Agent

   Copyright (C) 2002, 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/
/****************************************************************************
 *    INTERFACE REQUIRED HEADERS
 ****************************************************************************/

#include <time.h>

#include "ecurses.h"
#include "address.h"
#include "raddress.h"
#include "rstring.h"
#include "rmime.h"

/****************************************************************************
 *    INTERFACE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define FLAG_READ       0x0001
#define FLAG_OLD        0x0002
#define FLAG_ANSWERED   0x0004
#define FLAG_TRASHED    0x0008
#define FLAG_DRAFT      0x0010
#define FLAG_FLAGGED    0x0020
#define FLAG_PASSED     0x0040
#define FLAG_LEGITIMATE 0x0080
#define FLAG_NOSPAM     0x0100
#define FLAG_SPAM       0x0200
#define FLAG_MAILDIR    0x8000  /* set when mail comes from maildir, not set
                                   when mail comes from mbox */

/*******************************************
 * WHEN CHANGING SOMETHING BELOW:
 *   Remember to update mua_names in mail.c
 *******************************************/
typedef enum {
  MUA_UNKNOWN,
  
  /*----------- CONSOLE */
  MUA_ELMO,            /* User-Agent: elmo...*/
  MUA_MUTT,            /* User-Agent: Mutt...*/
  MUA_GNUS,            /* User-Ageng: Gnus...*/
  MUA_PINE,            /* Message-ID: PINE...*/
  MUA_ELM,             /* X-Mailer: ELM ...*/

  /*----- USER-FRIENDLY */
  MUA_KMAIL,           /* X-Mailer: KMail ...*/
  MUA_EVOLUTION,       /* X-Mailer: (Ximian)? Evolution...*/
  MUA_SYLPHEED,        /* X-Mailer: Sylpheed...*/

  /*----------- WINDOWS */
  MUA_MOZILLA,         /* User-Agent: Mozilla... */
  MUA_THE_BAT,         /* X-Mailer: The Bat!... */
  MUA_EUDORA,          /* X-Mailer: ... Eudora ... */
  MUA_OUTLOOK_EXPRESS, /* X-Mailer: Microsoft Outlook Express... */
  MUA_OUTLOOK,         /* X-Mailer: Microsoft Outlook IMO... */
  
  /*--------------- WWW */
  MUA_WWW_WEBMAIL,     /* X-Mailer: ... WebMail... */
  MUA_WWW_SQUIRREL,    /* X-Mailer: SquirellMail... */
  MUA_WWW_WP,          /* X-Mailer: Interfejs WWW poczty Wirtualnej Polski */
  MUA_WWW_ONET,        /* X-Mailer: onet.poczta */
  MUA_WWW_HOTMAIL,     /* X-OriginalArrivalTime: ... */
} mua_t;


#define MUA_CONSOLE(a) ((a) <= MUA_ELM)
#define MUA_NON_WIN(a) ((a) <= MUA_SYLPHEED)
#define MUA_WWW(a)     ((a) >= MUA_WWW_WEBMAIL)

typedef unsigned short flag_t;

/****************************************************************************
 *    INTERFACE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    INTERFACE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/

/*
 * This simple structure describes a single mail and its connections with
 * other mails grouped in mail_array, and in a tree which is built when
 * grouping messages in threads.
 * These information are only essentials taken from RFC822 and RFC2822,
 * and do not include any extension.
 *
 * parent      -- if this mail is not a reply -1, mail_array index otherwise
 * from        -- self-commenting
 * subject     -- self-commenting
 * date_str    -- date from header (verbatim)
 * to          -- self-commenting
 * reply_to    -- self-commenting
 * msg_id      -- Message-ID
 * in_reply_to -- In-Reply-To / References
 * tree        -- thread tree without colour
 * tree_len    -- tree length
 * index       -- index in mail_array
 * flags       -- bit flags (explained above)
 * date        -- local date
 * child_count -- reply count (this does not include replies to reply)
 * mua         -- Mail User Agent (see above)
 *
 * file_name     -- message file name (maildir)
 * offset_header -- byte offset of beginning of message header in file (mbox)
 * offset_body   -- byte offset of beginning of message body in file
 * offset_end    -- byte offset of end of message in file
 */
struct mail_st {

  int            parent;
  address_t     *from;
  char          *subject;
  char          *date_str;
  raddress_t    *to;
  raddress_t    *cc;
  raddress_t    *bcc;
  address_t     *reply_to;
  char          *msg_id;
  rstring_t     *in_reply_to;
  chtype        *tree;
  unsigned       tree_len;
  short          index;
  flag_t         flags;
  char           child_count;
  time_t         date;
  mua_t          mua;
  rmime_t       *rmime;
  
  union {
    unsigned  offset_header;
    char     *file_name;
  } place;
};

typedef struct mail_st mail_t;

/****************************************************************************
 *    INTERFACE DATA DECLARATIONS
 ****************************************************************************/

extern mail_t   *mail_array;
extern unsigned  mail_count;
extern unsigned  mail_not_replies_count;
extern unsigned  mail_read_count;
extern unsigned  mail_old_count;

extern char     *mail_mua_names[];

/****************************************************************************
 *    INTERFACE FUNCTION PROTOTYPES
 ****************************************************************************/

extern void    mail_destroy (mail_t *mail);
extern void    mail_free_resources (void);
extern void    mail_reset_array (void);
extern void    mail_set_array_size (size_t size);
extern mail_t *mail_insert (mail_t *mail);
extern void    mail_find_parents (void);
extern mail_t *mail_new_array (void);
extern void    mail_switch_to_array (mail_t *array);
extern void    mail_copy_to_array (mail_t *array, unsigned oldindex,
                                   unsigned newindex);

/****************************************************************************
 *    INTERFACE OBJECT CLASS DEFINITIONS
 ****************************************************************************/
/****************************************************************************
 *    INTERFACE TRAILING HEADERS
 ****************************************************************************/
/****************************************************************************
 *
 *    END HEADER mail.h
 *
 ****************************************************************************/
#endif

