/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2002, 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   These are all primitives I need to use when connecting as client.
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "xmalloc.h"
#include "error.h"
#include "networking.h"
#include "gettext.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define RECV_BUFFER_SIZE 1101
#define SEND_BUFFER_SIZE 1001

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/

extern void dump_buffer (char *buffer);

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static int connected = 0;

static struct in_addr server_address;
static unsigned short server_port;
static int            sock;

static char   *recv_buffer   = NULL;
static char   *send_buffer   = NULL;
static size_t  recv_buf_size = 0;
static size_t  send_buf_size = 0;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/

struct in_addr net_local_address;
unsigned short net_local_port;

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/

static int resolve_host (const char *hostname, unsigned short port);
static int make_connection (void);

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static int
resolve_host (const char *hostname, unsigned short port){
  struct hostent     *ht;

  error_msg (_("resolving host %s..."), hostname);
  ht = gethostbyname (hostname);
  if (!ht){
    error_ (0, 0, "%s: %s", hostname, hstrerror (h_errno));
    return 1;
  }
  server_address = *(struct in_addr *)ht->h_addr_list[0];
  server_port    = port;
  return 0;
}


static int
make_connection (void){
  struct sockaddr_in  st;
  socklen_t           st_len;

  error_msg (_("connecting to %s..."), inet_ntoa (server_address));
  sock = socket (PF_INET, SOCK_STREAM, 0);
  if (sock == -1){
    error_ (0, errno, "socket");
    return 1;
  }

  st.sin_family = AF_INET;
  st.sin_addr   = server_address;
  st.sin_port   = htons (server_port);

  if (connect (sock, (struct sockaddr *)&st, sizeof (st)) == -1){
    error_ (0, errno, "connect");
    return 1;
  }

  if (getsockname (sock, (struct sockaddr *)&st, &st_len) == 0){
    net_local_address = st.sin_addr;
    net_local_port    = st.sin_port;
  }
  return 0;
}



static void
allocate_resources (void)
{
  if (send_buf_size == 0){
    send_buf_size = SEND_BUFFER_SIZE;
    send_buffer   = xmalloc (send_buf_size);
  }
  if (recv_buf_size == 0){
    recv_buf_size = RECV_BUFFER_SIZE;
    recv_buffer   = xmalloc (recv_buf_size);
  }
}



static int
send_format_msg (const char *fmt, va_list ap)
{
  int ret;
  
  if (fmt && *fmt){
    ret = vsnprintf (send_buffer, send_buf_size - 1, fmt, ap);
    while (ret == -1 || ret >= send_buf_size - 1){
      send_buf_size = (send_buf_size + 1) * 2;
      send_buffer   = xrealloc (send_buffer, send_buf_size);
      ret           = vsnprintf (send_buffer, send_buf_size - 1, fmt, ap);
    }

    if (net_send (send_buffer, ret) == -1){
      net_close_connection ();
      error_ (0, errno, _("send"));
      return 1;
    }
  }
  return 0;
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

void
net_free_resources (void)
{
  if (recv_buf_size){
    xfree (recv_buffer);
    recv_buf_size = 0;
    recv_buffer   = NULL;
  }
  if (send_buf_size){
    xfree (send_buffer);
    send_buf_size = 0;
    send_buffer   = NULL;
  }
}


int
net_open_connection (const char *hostname, unsigned short port){

  if (resolve_host (hostname, port))
    return 1;

  if (make_connection ())
    return 1;

  connected = 1;
  return 0;
}



void
net_close_connection (void){

  if (!connected){
    errno = 0;
    return;
  }

  shutdown (sock, 2);
  connected = 0;
}



int
net_send (const void *msg, size_t len){

  if (!connected){
    errno = 0;
    return -1;
  }
  if (len == 0){
    return 0;
  }
  return send (sock, msg, len, 0);
}



int
net_recv (void *buf, size_t len){

  if (!connected){
    errno = 0;
    return -1;
  }
  return recv (sock, buf, len, 0);
}



int
net_send_txt (const char *buf){
  size_t len = strlen (buf);

  if (!connected){
    errno = 0;
    return -1;
  }
  return send (sock, buf, len, 0);
}



int
net_exchange_msg (int (*parse_fun)(char *),
                  int (*is_complete)(char *, size_t),
                  char *fmt, ...){

  int     ret;
  off_t   offset = 0;
  va_list ap;

  if (!connected){
    errno = 0;
    return 1;
  }
  
  allocate_resources ();
  
  va_start (ap, fmt);
  ret = send_format_msg (fmt, ap);
  va_end (ap);
  if (ret)
    return ret;
  
  do {
    ret = net_recv (recv_buffer + offset, recv_buf_size - 1 - offset);
    if (ret == -1){
      net_close_connection ();
      error_ (0, errno, "recv");
      return 1;
    }
    if (offset == recv_buf_size - 1){
      recv_buf_size = (recv_buf_size + 1) * 2;
      recv_buffer   = xrealloc (recv_buffer, recv_buf_size);
    }
    offset             += ret;
    recv_buffer[offset] = '\0';
  }
  while (!is_complete (recv_buffer, offset));

  return parse_fun (recv_buffer);
}


/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE networking.c
 *
 ****************************************************************************/
