/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2002 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   This module implements RFC1725 pop3 client side protocol.
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "networking.h"
#include "rstring.h"
#include "error.h"
#include "pop.h"
#include "xmalloc.h"
#include "mlex.h"
#include "progress.h"
#include "gettext.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static FILE *write_fp = NULL;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/

unsigned pop_mail_count = 0;
unsigned pop_mail_size  = 0;

pop_stat_t *pop_stat_list = NULL;
size_t      pop_stat_size = 0;

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/

static int  parse_function (char *buf);
static int  parse_stat (char *buf);
static void fill_stat_item (pop_stat_t *item, char *txt);
static int  is_complete (char *buf, size_t len);
static int  is_complete_multiline (char *buf, size_t len);

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static int
parse_function (char *buf)
{

  switch (*buf){
  case '+':
    return 0;

  case '-':
    return 1;

  default:
    return 1;
  }
}



/*
 * I assume that strtol returns 0 in case of error.
 */
static int
parse_stat (char *buf)
{
  int   ret;
  char *seek;
  char *rest;

  ret = parse_function (buf);
  if (ret)
    return ret;

  seek = strchr (buf, ' ');
  if (!seek)
    return 1;

  pop_mail_count = strtol (seek + 1, &rest, 10);
  if (*rest != ' ')
    return 1;

  pop_mail_size = strtol (rest + 1, &seek, 10);
  if (*seek)
    return 1;

  return 0;
}



static int
parse_list (char *buf)
{
  rstring_t  *lines;
  int         i;
  int         ret;
  
  ret = parse_function (buf);
  if (ret)
    return ret;
  
  lines = rstring_split_re (buf, "\r?\n");

  /* There is "+OK" in first line...*/
  pop_free_resources ();
  pop_stat_size = lines->count - 1;
  pop_stat_list = xmalloc (pop_stat_size * sizeof (pop_stat_t));
  
  for (i = 0; i < lines->count - 1; i++){
    fill_stat_item (pop_stat_list + i, lines->array[i + 1]);
  }
  rstring_delete (lines);
  
  return 0;
}



static int
parse_top (char *buf)
{
  int   ret;
  char *seek;

  ret = parse_function (buf);
  if (ret)
    return ret;

  seek = strstr (buf, "\r\n");
  if (!seek)
    seek = buf;
  else
    seek += 2;

  ret = mlex_scan_buffer (seek);
  if (ret == NEXT_MAIL)
    return 0;
  return 1;
}



static int
parse_retr (char *buf)
{
  char *seek;
  char *tmp;
  int   ret;

  ret = parse_function (buf);
  if (ret)
    return 1;
  
  if (write_fp == NULL)
    return 1;

  /**
   * step 0: omit first line, its like "+OK...\r\n"
   */
  seek = strstr (buf, "\r\n");
  if (seek == NULL)
    seek = buf;
  else
    seek += 2;

  /**
   * step 1: write mail char by char, omit \r and leading dot
   */
  while (1){
    tmp = strstr (seek, "\r\n");
    if (tmp == NULL)
      break;
    if (*seek == '.')
      seek++;
    fwrite (seek, 1, tmp - seek, write_fp);
    fputc ('\n', write_fp);
    seek = tmp + 2;
  }
  fprintf (write_fp, "%s\n", seek);
  
  write_fp = NULL;
  return 0;
}



static void
fill_stat_item (pop_stat_t *item, char *txt)
{
  char *rest;

  item->num     = strtol (txt, &rest, 10);
  item->size    = strtol (rest, NULL, 10);
  item->mark    = 0;
  item->fetched = 0;
  item->info    = NULL;
}



static void
refill_stat_item (pop_stat_t *item)
{
  item->info = xmalloc (sizeof (mail_t));
  memcpy (item->info, newmail, sizeof (mail_t));
}



static int
is_complete (char *buf, size_t len)
{
  if (buf[len - 1] == '\n' && buf[len - 2] == '\r'){
    buf[len - 2] = '\0';
    return 1;
  }
  else {
    return 0;
  }
}



static int
is_complete_multiline (char *buf, size_t len)
{
  if (*buf == '-')
    return is_complete (buf, len);
  
  if (len          >= 5                                                   &&
      buf[len - 1] == '\n' && buf[len - 2] == '\r' && buf[len - 3] == '.' &&
      buf[len - 4] == '\n' && buf[len - 5] == '\r'){
    buf[len - 3]    = '\0';
    return 1;
  }
  else {
    return 0;
  }
}



static void
empty_stat_list (void)
{
  int i;

  for (i = 0; i < pop_stat_size; i++){
    if (pop_stat_list[i].info){
      mail_destroy (pop_stat_list[i].info);
      xfree (pop_stat_list[i].info);
    }
  }
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/


void
pop_free_resources (void)
{
  if (pop_stat_list){
    empty_stat_list ();
    xfree (pop_stat_list);
    pop_stat_list = 0;
    pop_stat_size = 0;
  }
}
    



int
pop_begin_connection (char *hostname, unsigned short port)
{
  port = (port) ? port : 110;
  
  if (net_open_connection (hostname, port)){
    return 1;
  }

  return net_exchange_msg (parse_function, is_complete, NULL);
}



void
pop_finish_connection (void)
{
  net_exchange_msg (parse_function, is_complete, "QUIT\r\n");
  net_close_connection ();
}



int
pop_authorize (const char *user, const char *pass)
{
  error_msg (_("logging as %s..."), user);

  if (net_exchange_msg (parse_function, is_complete, "USER %s\r\n", user)){
    error_ (0, 0, _("invalid user name %s"), user);
    return 1;
  }

  if (net_exchange_msg (parse_function, is_complete, "PASS %s\r\n", pass)){
    error_ (0, 0, _("password %s rejected"), pass);
    return 1;
  }

  error_clear_force ();
  return 0;
}



int
pop_rset (void)
{
  return net_exchange_msg (parse_function, is_complete, "RSET\r\n");
}



int
pop_stat (void)
{
  return net_exchange_msg (parse_stat, is_complete, "STAT\r\n");
}



int
pop_list (void)
{
  return net_exchange_msg (parse_list, is_complete_multiline, "LIST\r\n");
}


int
pop_list_ext (void)
{
  int ret;
  int i;
  
  progress_open (_("retrieving list"));
  progress_set_max (pop_stat_size);
  
  for (i = 0; i < pop_stat_size; i++){
    if (pop_stat_list[i].mark == 1)
      continue;
    ret = net_exchange_msg (parse_top, is_complete_multiline, "TOP %d 0\r\n",
                            pop_stat_list[i].num);
    refill_stat_item (pop_stat_list + i);
    progress_add (1);
  }

  progress_close ();

  return 0;
}


int
pop_dele (int num)
{
  return net_exchange_msg (parse_function, is_complete, "DELE %d\r\n", num);
}


int
pop_retr (int index, FILE *fp)
{
  write_fp = fp;
  
  return net_exchange_msg (parse_retr, is_complete_multiline, "RETR %d\r\n",
                           index);
}


/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE pop.c
 *
 ****************************************************************************/
