/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   selection window
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <string.h>

#include "ecurses.h"
#include "select.h"
#include "xmalloc.h"
#include "screen.h"
#include "error.h"
#include "line.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))


#define PREAMBLE do {if (select == NULL) return; } while (0)

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/
/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static void
validate (select_t *select)
{
  int position;
  
  if (select->no_bar)
    return;
  
  if (select->top_pos >= select->items->count)
    select->top_pos = select->items->count - 1;
  if (select->top_pos < 0)
    select->top_pos = 0;
  if (select->bar_pos >= select->items->count)
    select->bar_pos = select->items->count - 1;
  if (select->bar_pos < 0)
    select->bar_pos = 0;

  position = select->bar_pos - select->top_pos;

  mvwin (select->bar, select->win->_begy + position + select->with_box,
         select->bar->_begx);
}


static void
put_items (select_t *select)
{
  int       i;
  unsigned  index;
  line_t   *line;
  
  validate (select);
  index = select->top_pos;
  line  = line_create (select->win->_maxx + 1 - 2 * select->with_box);
  
  for (i = 0; i < select->win->_maxy - select->with_box
         && index + i < select->items->count; i++){
    line_put_str (line, select->items->array[index + i]);
    mvwaddstr (select->win, i + select->with_box, select->with_box, line->str);
  }
  line_clear (line);
  while (i < select->win->_maxy - select->with_box){
    mvwaddstr (select->win, i + select->with_box, select->with_box, line->str);
    i++;
  }
  line_destroy (line);
}


static void
draw_bar (select_t *select)
{
  line_t *line;

  if (select->no_bar)
    return;
  
  line = line_create (select->win->_maxx + 1 - 2 * select->with_box);

  if (select->items->count){
    line_put_str (line, select->items->array[select->bar_pos]);
  }
  else {
    line_clear (line);
  }
  mvwaddstr (select->bar, 0, 0, line->str);
  line_destroy (line);
}



static int
bar_on_top (select_t *select)
{
  if (select->no_bar)
    return 1;
  
  return select->bar_pos == select->top_pos;
}


static int
bar_on_bottom (select_t *select)
{
  if (select->no_bar)
    return 1;
  
  return select->bar_pos == select->top_pos + select->win->_maxy
    - 2 * select->with_box;
}


static int
bar_on_first_page (select_t *select)
{
  if (select->no_bar)
    select->bar_pos = select->top_pos;
  return select->bar_pos <= select->win->_maxy - 2 * select->with_box;
}


static int
bar_on_last_page (select_t *select)
{
  if (select->no_bar)
    select->bar_pos = select->top_pos;
  return select->bar_pos + select->win->_maxy >= select->items->count;
}


static void
scroll_down (select_t *select)
{
  select->top_pos++;
  put_items (select);
}


static void
scroll_up (select_t *select)
{
  select->top_pos--;
  put_items (select);
}


static void
refresh_me (select_t *select)
{
  screen_touchwin (select->win);
  screen_wnoutrefresh (select->win);
  if (!select->no_bar){
    screen_touchwin (select->bar);
    screen_wnoutrefresh (select->bar);
  }
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

select_t *
select_open (rstring_t *items, int w, int h, int with_box, int no_bar,
             const char *title)
{
  select_t *result;

  if (items == NULL)
    return NULL;
  
  if (items->count > h - 2 * with_box && items->count < LINES - 8)
    h = items->count + 2 * with_box;
  
  result           = xmalloc (sizeof (select_t));
  result->items    = items;
  result->top_pos  = 0;
  result->bar_pos  = 0;
  result->with_box = with_box;
  result->no_bar   = no_bar;
  result->win      = newwin (h, w, (LINES - h) / 2, (COLS - w) / 2);
  if (!no_bar){
    result->bar    = newwin (1, w - 2 * with_box,
                             result->win->_begy + with_box,
                             result->win->_begx + with_box);
    wbkgdset (result->bar, COLOR_PAIR (2));
  }

  wattron (result->win, A_BOLD);
  if (with_box)
    box (result->win, 0, 0);

  if (title){
    mvwaddstr (result->win, 0, (w - strlen (title)) / 2, title);
  }
  wattroff (result->win, A_BOLD);

  put_items (result);
  if (items->count)
    draw_bar (result);

  return result;
}




void
select_move_vert (select_t *select, int newy)
{
  PREAMBLE;
  mvwin (select->win, newy, select->win->_begx);
}



int
select_close (select_t *select)
{
  int result;

  if (select == NULL)
    return -1;

  result = select->bar_pos;
  
  delwin (select->win);
  if (!select->no_bar)
    delwin (select->bar);

  xfree (select);

  return result;
}



void
select_show (select_t *select)
{
  PREAMBLE;
  refresh_me (select);
}



void
select_next (select_t *select)
{
  PREAMBLE;
  if (select->bar_pos >= select->items->count - 1)
    return;

  if (bar_on_bottom (select)){
    select->bar_pos++;
    scroll_down (select);
    draw_bar (select);
  }
  else {
    select->bar_pos++;
    mvwin (select->bar, select->bar->_begy + 1, select->bar->_begx);
    draw_bar (select);
  }
  refresh_me (select);
}



void
select_prev (select_t *select)
{
  PREAMBLE;
  if (select->bar_pos == 0)
    return;

  if (bar_on_top (select)){
    select->bar_pos--;
    scroll_up (select);
    draw_bar (select);
  }
  else {
    select->bar_pos--;
    mvwin (select->bar, select->bar->_begy - 1, select->bar->_begx);
    draw_bar (select);
  }
  refresh_me (select);
}



void
select_next_page (select_t *select)
{
  PREAMBLE;
  if (bar_on_last_page (select)){
    select_last (select);
    return;
  }

  select->bar_pos += select->win->_maxy - 2 * select->with_box;
  select->top_pos += select->win->_maxy - 2 * select->with_box;
  select_redraw (select);
}



void
select_prev_page (select_t *select)
{
  PREAMBLE;
  if (bar_on_first_page (select)){
    select_first (select);
    return;
  }

  select->bar_pos -= select->win->_maxy - 2 * select->with_box;
  select->top_pos -= select->win->_maxy - 2 * select->with_box;
  select_redraw (select);
}



void
select_first (select_t *select)
{
  PREAMBLE;
  select->bar_pos = 0;
  select->top_pos = 0;
  select_redraw (select);
}



void
select_last (select_t *select)
{
  int pos;

  PREAMBLE;

  pos = select->items->count - 1 - select->win->_maxy
    + 2 * select->with_box;
  
  if (select->top_pos < pos)
    select->top_pos = pos;
  if (select->no_bar)
    select->bar_pos = select->top_pos;
  else
    select->bar_pos = select->items->count - 1;
  select_redraw (select);
}


void
select_redraw (select_t *select)
{
  PREAMBLE;
  put_items (select);
  draw_bar (select);
  refresh_me (select);
}



void
select_add_item (select_t *select, char *item)
{
  PREAMBLE;
  rstring_add (select->items, item);
  select_redraw (select);
}



void
select_remove_item (select_t *select, int index)
{
  PREAMBLE;
  rstring_remove (select->items, index);
  select_redraw (select);
}



void
select_remove_selected (select_t *select)
{
  PREAMBLE;
  select_remove_item (select, select->bar_pos);
}



void
select_install_new_items (select_t *select, rstring_t *items)
{
  PREAMBLE;
  rstring_delete (select->items);
  select->items = items;
  select_redraw (select);
}


/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE select.c
 *
 ****************************************************************************/
