/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------


   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "smtp.h"
#include "clock.h"
#include "ask.h"
#include "error.h"
#include "eprintf.h"
#include "screen.h"
#include "rstring.h"
#include "mlex.h"
#include "select.h"
#include "file.h"
#include "xmalloc.h"
#include "mailinfo.h"
#include "read.h"
#include "compose.h"
#include "line.h"
#include "misc.h"
#include "cmd.h"
#include "gettext.h"
#include "wrapbox.h"
#include "mybox.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

typedef enum {SENDER_DROP, SENDER_DRAFT, SENDER_SENT} sent_t;

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static select_t *sender_select = NULL;
static mail_t   *message       = NULL;

static line_t   *line          = NULL;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/

int sender_successfull_send = 0;

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static int
parse_mail (char *file)
{
  yyin = fopen (file, "r");
  if (yyin == NULL){
    error_ (0, errno, file);
    return 1;
  }

  if (mlex_outmail_scan () != NEXT_MAIL){
    error_ (0, 0, _("file %s has bad format"), file);
    return 1;
  }
  fclose (yyin);

  message = xmalloc (sizeof (mail_t));

  memcpy (message, newmail, sizeof (mail_t));

  message->index           = -1;
  message->flags           = FLAG_MAILDIR;
  message->place.file_name = file;

  return 0;
}



static int
write_headers (FILE *fp, mime_t *mime, char *boundary)
{
  char *fname;
  
  if (boundary == NULL)
    return 0;

  fprintf (fp, "--%s\n", boundary);
  if (mime->type_desc)
    fprintf (fp, "Content-Type: %s\n", mime->type_desc);
  else
    fprintf (fp, "Content-Type: %s\n", mime_type_str[mime->type]);

  if (mime->file_name){
    fname = strrchr (mime->file_name, '/');
    if (fname == NULL)
      fname = mime->file_name;
    else
      fname++;
    fprintf (fp, "Content-Disposition: attachment; filename=\"%s\"\n", fname);
  }
  else {
    fprintf (fp, "Content-Disposition: inline\n");
  }

  fprintf (fp, "Content-Transfer-Encoding: %s\n",
           mime_enc_str[mime->encoding]);
  fprintf (fp, "\n");

  return 0;
}



static int
write_attachment (FILE *fp, mime_t *mime)
{
  char *body;
  int  len;

  body = mime_get_and_encode (mime, &len, message->place.file_name);

  fwrite (body, 1, len, fp);
  xfree (body);

  return 0;
}



static int
write_end (FILE *fp, char *boundary)
{
  if (boundary)
    fprintf (fp, "--%s--\n", boundary);
  return 0;
}



static void
write_references (FILE *fp, rstring_t *references)
{
  char **iterator;
  
  fprintf (fp, "References:");
  for (iterator = references->array; *iterator; iterator++){
    fprintf (fp, " %s", *iterator);
  }
  fprintf (fp, "\n");
  fprintf (fp, "In-Reply-To: %s\n", references->array[references->count - 1]);
}



static int
rewrite_file (void)
{
  FILE   *fp;
  char   *tmp_file_name;
  char   *boundary = NULL;
  int     i;

  fp = file_temp_file (&tmp_file_name);
  if (fp == NULL){
    error_ (0, errno, tmp_file_name);
    return 1;
  }
  if (message->from && message->from->full)
    fprintf (fp, "From: %s\n", message->from->full);
  if (message->to && message->to->count){
    fprintf (fp, "To: %s", message->to->array[0]->full);
    for (i = 1; i < message->to->count; i++){
      fprintf (fp, ",\n\t%s", message->to->array[i]->full);
    }
    fprintf (fp, "\n");
  }
  if (message->subject)
    fprintf (fp, "Subject: %s\n", message->subject);
  if (message->date_str)
    fprintf (fp, "Date: %s\n", message->date_str);
  if (message->msg_id)
    fprintf (fp, "Message-ID: %s\n", message->msg_id);
  if (message->in_reply_to)
    write_references (fp, message->in_reply_to);
  fprintf (fp, "User-Agent: %s/%s\n", PACKAGE, VERSION);
  fprintf (fp, "MIME-Version: 1.0\n");

  if (message->rmime->count == 1){
    fprintf (fp, "Content-Type: text/plain; charset=iso-8859-2\n");
    fprintf (fp, "Content-Transfer-Encoding: 8bit\n");
    fprintf (fp, "\n");
  }
  else {
    boundary = compose_boundary ();
    fprintf (fp, "Content-Type: multipart/mixed; boundary=\"%s\"\n", boundary);
    fprintf (fp, "\n");
  }

  for (i = 0; i < message->rmime->count; i++){
    write_headers (fp, message->rmime->array[i], boundary);
    write_attachment (fp, message->rmime->array[i]);
  }
  write_end (fp, boundary);

  fclose (fp);
  if (boundary)
    xfree (boundary);
  
  unlink (message->place.file_name);
  xfree (message->place.file_name);
  message->place.file_name = tmp_file_name;
  return 0;
}



static char *
get_fname (mime_t *mime)
{
  return (mime->file_name) ? mime->file_name : message->place.file_name;
}



/**
 *    <= 14     <= 6  <= 10
 * [text/plain, 8bit, 8.0K]
 */
static char *
get_info (mime_t *mime)
{
  char        *result = xmalloc (1 + 14 + 6 + 10 + 1 + 1);
  char        *seek;
  static char  type[15];
  static char  enc[7];

  type[14] = '\0';
  enc[6]   = '\0';

  strncpy (enc, mime_enc_str[mime->encoding], 6);
  if (mime->type_desc){
    strncpy (type, mime->type_desc, 14);
  }
  else {
    strncpy (type, mime_type_str[mime->type], 14);
  }
  seek = strchr (type, ';');
  if (seek)
    *seek = '\0';


  sprintf (result, "[%s, %s, %s]", type, enc,
           human_size_pure (mime->offset_end - mime->offset_start));

  return result;
}




static char *
item_text (mime_t *mime)
{
  char *fname = get_fname (mime);
  char *info  = get_info (mime);
  
  if (line == NULL)
    line = line_create (COLS + 1);

  line_put_str (line, fname);
  line_add_str_end (line, info);

  xfree (info);

  return xstrdup (line->str);
}




static void
leave_window (sent_t sent)
{
  char *save_box = NULL;
  
  if (message){

    if (sent == SENDER_DRAFT)
      save_box = mybox_draft ();
    else if (sent == SENDER_SENT)
      save_box = mybox_sent ();
    
    if (save_box){
      wrapbox_move_mail_to (message, save_box);
      xfree (save_box);
    }
    else {
      unlink (message->place.file_name);
    }
    mail_destroy (message);
    xfree (message);
    message = NULL;
  }
  if (line){
    line_destroy (line);
    line = NULL;
  }
  rstring_free_strings (sender_select->items);
  rstring_delete (sender_select->items);
  select_close (sender_select);
  mailinfo_close ();
}



/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/


int
sender_open (char *file)
{
  rstring_t *items;

  if (parse_mail (file))
    return 1;
  
  items = rstring_create_size (2);
  rstring_add (items, item_text (message->rmime->array[0]));
  
  sender_select = select_open (items, COLS, LINES - MAILINFO_LINES - 2, 0, 0,
                               NULL);
  select_move_vert (sender_select, MAILINFO_LINES);
  select_redraw (sender_select);

  mailinfo_show (message);
  return 0;
}




void
sender_refresh (void)
{
  mailinfo_refresh ();
  select_redraw (sender_select);
}



void
sender_open_new (void)
{
  char *file = compose_new ();

  if (sender_open (file))
    return;

  cmd_state_push (CMD_SENDER);
}



void
sender_open_reply (void)
{
  char *file = compose_reply ();

  if (sender_open (file))
    return;

  cmd_state_push (CMD_SENDER);
}



void
sender_open_fwd (void)
{
  char *file = compose_fwd ();

  if (sender_open (file))
    return;

  cmd_state_push (CMD_SENDER);
}



void
sender_close (void)
{
  if (ask_if_sure ("save message as draft (y/n)? ")){
    leave_window (SENDER_DRAFT);
  }
  else {
    leave_window (SENDER_DROP);
  }
  
  sender_successfull_send = 0;
  cmd_state_pop ();
}



void
sender_next (void)
{
  select_next (sender_select);
}



void
sender_prev (void)
{
  select_prev (sender_select);
}



void
sender_next_page (void)
{
  select_next_page (sender_select);
}



void
sender_prev_page (void)
{
  select_prev_page (sender_select);
}



void
sender_first (void)
{
  select_first (sender_select);
}



void
sender_last (void)
{
  select_last (sender_select);
}



void
sender_change_from (void)
{
  char *from = read_argument ("from: ", message->from->full);

  if (from){
    message->reply_to = message->from = address_create (from);
  }

  mailinfo_refresh ();
}



void
sender_change_reply_to (void)
{
  char *reply_to = read_argument ("reply-to: ", message->reply_to->full);

  if (reply_to){
    message->reply_to = address_create (reply_to);
  }

  mailinfo_refresh ();
}



void
sender_add_attachment (void)
{
  mime_t *mime;
  char   *attachment;

  attachment = read_argument ("file: ", NULL);
  if (attachment && *attachment){
    if (file_size (attachment) == -1){
      error_ (0, errno, "%s", attachment);
      return;
    }
    mime = mime_create_from_filename (attachment);
    select_add_item (sender_select, item_text (mime));
    rmime_add (message->rmime, mime);
  }
}



void
sender_delete_attachment (void)
{
  mime_destroy (message->rmime->array[sender_select->bar_pos]);
  rmime_remove (message->rmime, sender_select->bar_pos);
  select_remove_selected (sender_select);
}



void
sender_change_subject (void)
{
  char *subject = read_argument ("subject: ", message->subject);

  if (subject && message->subject){
    xfree (message->subject);
    message->subject = xstrdup (subject);
  }
  else if (subject){
    message->subject = xstrdup (subject);
  }

  mailinfo_refresh ();
}



void
sender_change_content_type (void)
{
  mime_t *mime = message->rmime->array[sender_select->bar_pos];
  char   *type = read_argument ("Content-Type: ", mime->type_desc);
  char  **item = sender_select->items->array + sender_select->bar_pos;

  if (mime->type_desc){
    xfree (mime->type_desc);
    mime->type_desc = NULL;
  }
  if (type){
    mime_set_content_type (mime, type);
  }

  xfree (*item);
  *item = item_text (mime);
  select_redraw (sender_select);
}



void
sender_go (void)
{
  rstring_t *conf;
  char      *ask;
  char      *host;
  char      *copy;
  int        port = 0;

  ask = ask_value ("smtp_acc");

  if (ask == NULL)
    return;

  copy = xstrdup (ask);
  conf = rstring_split (copy, ":");
  host = conf->array[0];

  if (conf->count > 1)
    port = atoi (conf->array[1]);
  
  sender_successfull_send = 0;

  if (rewrite_file ()){
    return;
  }

  port = (port == -1) ? 0 : port;

  if (smtp_begin_connection (host, (unsigned short) port)){
    rstring_delete (conf);
    xfree (copy);
    return;
  }

  rstring_delete (conf);
  xfree (copy);

  if (smtp_ehlo (1)){
    smtp_finish_connection ();
    return;
  }

  if (smtp_mail_from (message->from->email)){
    smtp_finish_connection ();
    return;
  }

  if (smtp_send_mail (message->place.file_name, message->to)){
    smtp_finish_connection ();
    return;
  }

  smtp_finish_connection ();
  leave_window (SENDER_SENT);

  sender_successfull_send = 1;
  cmd_state_pop ();
}

void
sender_reedit (void)
{

  static char *command;
  char        *editor = ask_value ("editor");

  if (!editor || !*editor){
    editor = "jed %f";
  }

  command = eprintf_editor (editor, message->place.file_name, 1);

  curs_set (1);
  clock_stop ();
  endwin ();
  
  system (command);
  xfree (command);
  
  screen_refresh ();
  curs_set (0);
  clock_resume ();
//  folder_refresh ();
  mailinfo_refresh();
  sender_refresh();
}


/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE sender.c
 *
 ****************************************************************************/
